/*-----------------------------------------------------------------------------
  Project:  QRD
  File:     qrd.h

  Copyright (C)2010 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law. They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Revision: 1.7 $
 *----------------------------------------------------------------------------*/
#ifndef MW_QRD_QRD_H_
#define MW_QRD_QRD_H_

#include <nn/os.h>

#ifdef __cplusplus

namespace mw {
namespace qrd {

//==============================================================================

const int DATA_MAX = 7089;      //<! Maximum number of pieces of data that can be stored in a QR Code (version 40), according to its specifications

typedef struct tagQRDContext QRDContext;  // Module context structure

// Type of data stored in QR Code
typedef enum DataType
{
    DATA_TYPE_NUMERIC = 0,      //<! Numeric mode
    DATA_TYPE_ALPHABET,         //<! Alphanumeric mode
    DATA_TYPE_KANJI,            //<! Kanji mode (Shift_JIS encoding)
    DATA_TYPE_BINARY            //<! 8-bit binary data
} DataType;

// Error codes for QR Code decoding
typedef enum Result
{
    RESULT_SUCCESS = 0,         //<! Processing successful
    RESULT_ILLEGAL_PARAMETER,   //<! Illegal parameter
    RESULT_NODATA,              //<! Failed to recognize QR Code
    RESULT_NOMEMORY,            //<! Insufficient work buffer
    RESULT_VERSION,             //<! Unsupported version of QR Code
    RESULT_ILLEGAL_IMAGESIZE,   //<! Invalid image size
    RESULT_NOT_INITIALIZED      //<! Library not initialized
} Result;

// Structure storing QR Code position information
typedef struct Rect
{
    s32 top;
    s32 left;
    s32 right;
    s32 bottom;
} Rect;

// Structure that stores results of QR Code decoding
typedef struct DecodeData
{
    Rect qrPosition;            //<! QR Code position information
    s32 version;                //<! QR Code version number (1 to 40)
    s32 size;                   //<! Valid size of decoded data
    s32 total;                  //<! If the QR Code is divided, this is the number of divisions (0-15)
    s32 index;                  //<! Index (0-15) of divided QR Code
    s32 parity;                 //<! Parity of divided QR Code
    DataType type;              //<! Data type of decoded QR Code
    u8 data[DATA_MAX + 1];      //<! Decoded data
    NN_PADDING1;
} DecodeData;

// Structure that stores image to QR-decode
typedef struct ImageInfo
{
    u16 width;                  //<! Width of image to decode
    u16 height;                 //<! Height of image to decode
    u32* yuv422Data;            //<! Image data to decode (YUV422 format)
    Rect qrPosition;            //<! Rectangle to decode
} ImageInfo;

//==============================================================================

class QRDecoder
{
public:
    QRDecoder();
    virtual ~QRDecoder();

    bool Initialize(
        void* ptr,
        u32 size);

    void Finalize();

    Result ExecuteDecode(
        DecodeData* result,
        const ImageInfo* img);

    static u32 GetDecodeBufferSize(
        u16 width,
        u16 height);
};

} // namespace qre
} // namespcae mw

#endif // __cplusplus

//==============================================================================

#ifdef __cplusplus
# define MW_QRD_EXTERN_C extern "C"
#else
# define MW_QRD_EXTERN_C extern
#endif

/* Please see man pages for details 
 
 
 
 */
typedef union mwqrdQRDecoder
{
	char buf[4];
	u32 alignment_holder;
} mwqrdQRDecoder;

const int MW_QRD_DATA_MAX = 7089;      //<! Maximum number of pieces of data that can be stored in a QR Code (version 40), according to its specifications

// Type of data stored in QR Code
typedef enum mwqrdDataType
{
    MW_QRD_DATA_TYPE_NUMERIC = 0,      //<! Numeric mode
    MW_QRD_DATA_TYPE_ALPHABET,         //<! Alphanumeric mode
    MW_QRD_DATA_TYPE_KANJI,            //<! Kanji mode (Shift_JIS encoding)
    MW_QRD_DATA_TYPE_BINARY            //<! 8-bit binary data
} mwqrdDataType;

// Error codes for QR Code decoding
typedef enum mwqrdResult
{
    MW_QRD_RESULT_SUCCESS = 0,         //<! Processing successful
    MW_QRD_RESULT_ILLEGAL_PARAMETER,   //<! Illegal parameter
    MW_QRD_RESULT_NODATA,              //<! Failed to recognize QR Code
    MW_QRD_RESULT_NOMEMORY,            //<! Insufficient work buffer
    MW_QRD_RESULT_VERSION,             //<! Unsupported version of QR Code
    MW_QRD_RESULT_ILLEGAL_IMAGESIZE,   //<! Invalid image size
    MW_QRD_RESULT_NOT_INITIALIZED      //<! Library not initialized
} mwqrdResult;

// Structure storing QR Code position information
typedef struct mwqrdRect
{
    s32 top;
    s32 left;
    s32 right;
    s32 bottom;
} mwqrdRect;

// Structure that stores results of QR Code decoding
typedef struct mwqrdDecodeData
{
    mwqrdRect qrPosition;            //<! QR Code position information
    s32 version;                     //<! QR Code version number (1 to 40)
    s32 size;                        //<! Valid size of decoded data
    s32 total;                       //<! If the QR Code is divided, this is the number of divisions (0-15)
    s32 index;                       //<! Index (0-15) of divided QR Code
    s32 parity;                      //<! Parity of divided QR Code
    mwqrdDataType type;              //<! Data type of decoded QR Code
    u8 data[MW_QRD_DATA_MAX + 1];    //<! Decoded data
    NN_PADDING1;
} mwqrdDecodeData;

// Structure that stores image to QR-decode
typedef struct mwqrdImageInfo
{
    u16 width;                       //<! Width of image to decode
    u16 height;                      //<! Height of image to decode
    u32* yuv422Data;                 //<! Image data to decode (YUV422 format)
    mwqrdRect qrPosition;            //<! Rectangle to decode
} mwqrdImageInfo;

/* Please see man pages for details 
 
 */
MW_QRD_EXTERN_C bool mwqrdQRDecoderInitialize(mwqrdQRDecoder* this_, void* ptr, u32 size);

/* Please see man pages for details 
 
 */
MW_QRD_EXTERN_C void mwqrdQRDecoderFinalize(mwqrdQRDecoder* this_);

/* Please see man pages for details 
 
 */
MW_QRD_EXTERN_C mwqrdResult mwqrdQRDecoderExecuteDecode(mwqrdQRDecoder* this_, mwqrdDecodeData* result, const mwqrdImageInfo* img);

/* Please see man pages for details 
 
 */
MW_QRD_EXTERN_C u32 mwqrdQRDecoderGetDecodeBufferSize(u16 width, u16 height);

//==============================================================================

/* MW_QRD_QRD_H_ */
#endif
