/*---------------------------------------------------------------------------*
  Project:  QRD
  File:     GraphicsControl.cpp

  Copyright (C)2010 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: $
 *---------------------------------------------------------------------------*/

#include <nn/os.h>
#include <nn/gx.h>
#include <string.h>

#include "GraphicsControl.h"

/* ExpHeap for app. */
extern nn::fnd::ExpHeap s_AppHeap;

/* Please see man pages for details 
 
 
 */
int GraphicsControl::InitializeGraphics(void)
{
    s_HeapForGx = reinterpret_cast<uptr>(s_AppHeap.Allocate(0x800000));
    /* Initialize display */
    s_RenderSystem.Initialize(s_HeapForGx, 0x800000);

    s_RenderSystem.SetClearColor(NN_GX_DISPLAY0, 0.f, 0.f, 0.f, 1.0f);
    s_RenderSystem.SetClearColor(NN_GX_DISPLAY1, 0.f, 0.f, 0.f, 1.0f);

    glClearDepthf(1.f);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glFrontFace(GL_CCW);

    return 0;
}

/* Please see man pages for details 
 
 
 */
void GraphicsControl::FinalizeGraphics(void)
{
    /* shutdown_display */
    s_RenderSystem.Finalize();
    s_AppHeap.Free(reinterpret_cast<void*>(s_HeapForGx));
}

/* Please see man pages for details 
 
 
 */
void GraphicsControl::ClearDisplay(const GLenum display)
{
    s_RenderSystem.SetRenderTarget(display);
    s_RenderSystem.Clear();
}

/* Please see man pages for details 
 
 
 */
void GraphicsControl::SwapBuffer(const s32 display, const bool bWait)
{
    s_RenderSystem.SwapBuffers();
    if (bWait)
        s_RenderSystem.WaitVsync(display);
}

/* Please see man pages for details 
 
 
 */
void GraphicsControl::DrawLine(s32 display, f32 sx, f32 sy, f32 ex,
                               f32 ey, f32 w, f32 r, f32 g, f32 b)
{
    // s_RenderSystem.SetRenderingTarget(display);
    s_RenderSystem.SetRenderTarget(display);

    s_RenderSystem.SetColor(r, g, b);
    s_RenderSystem.SetLineWidth(w);
    // s_RenderSystem.DrawLine2d(sx, sy, ex, ey);
    s_RenderSystem.DrawLine(sx, sy, ex, ey);
}

/* Please see man pages for details 
 
 
 */
u32 GetTextureLength(const u32& imageLength)
{        
    u32 textureLength = 8;

    // 8, 16, 32, 64, 128, 256, 512, 1024
    for (u32 i = 0; i < 7; i++)
        {
            if (imageLength > textureLength)
                {
                    textureLength *= 2;
                }
            else
                {
                    return textureLength;
                }
        }

    return 1024;
}

/* Please see man pages for details 
 
 
 */
u8* GetTextureDataFromBmpFileData(u32& bmpWidth, u32& bmpHeight, u8* bmpDataBuffer, 
                                  u32& textureWidth, u32& textureHeight)
{
    u32 dispWidth = 320 < bmpWidth ? 320 : bmpWidth;
    u32 dispHeight = 240 < bmpHeight ? 240 : bmpHeight;
    // Convert BMP to OpenGL RGB format
    textureWidth = GetTextureLength(dispWidth);
    textureHeight = GetTextureLength(dispHeight);
    u8* textureGLDataBuffer = new GLubyte[3 * textureWidth * textureHeight];    
    for (u32 y = 0; y < textureHeight; y++)
        {
            for (u32 x = 0; x < textureWidth; x++)
                {                        
                    u8* textureDataPtr = textureGLDataBuffer;
                    textureDataPtr += 3 * ((static_cast<u32>(textureWidth) * y) + x);
                    if ((x >= dispWidth) || (y >= dispHeight))
                        {
                            (*textureDataPtr) = 0x00;
                            textureDataPtr += 1;
                
                            (*textureDataPtr) = 0x00;
                            textureDataPtr += 1;
                
                            (*textureDataPtr) = 0x00;
                        }
                    else
                        {
                            u8* bmpDataPtr = bmpDataBuffer;

                            bmpDataPtr += 3 * (bmpWidth * (bmpHeight - y * 2 - 2) + x * 2);
                            (*textureDataPtr) = (*(bmpDataPtr + 2));
                            textureDataPtr += 1;
                            (*textureDataPtr) = (*(bmpDataPtr + 1));
                            textureDataPtr += 1;
                
                            (*textureDataPtr) = (*(bmpDataPtr + 0));
                        }
                }
        }

    u8* textureDataBuffer = reinterpret_cast<u8*>(s_AppHeap.Allocate(3 * textureWidth
                                                                     * textureHeight));
    // Convert OpenGL RGB format to PICA Native RGB format
    GLenum format = GL_RGB_NATIVE_DMP;
    bool result = demo::ConvertGLTextureToNative(format, textureWidth, textureHeight,
                                                 textureGLDataBuffer, textureDataBuffer);
    if (!result)
        {
            NN_LOG("  Convert to GL_RGB_NATIVE_DMP failed.\n");
            NN_ASSERT(0);
        }

    delete [] textureGLDataBuffer;

    bmpWidth = dispWidth;
    bmpHeight = dispHeight;

    return textureDataBuffer;
}

/* Please see man pages for details 
 
 
 */
void GraphicsControl::DrawCamera(u8* textureDataPtr, u32 bmpWidth, u32 bmpHeight)
{
    static GLuint textureId    = 0;
    if (textureId != 0)
        {
            s_RenderSystem.DeleteTexture(textureId);
        }

    /// RGB 24Bit
    GLenum target         = GL_TEXTURE_2D;
    GLenum internalFormat = GL_RGB_NATIVE_DMP;
    GLenum format         = GL_RGB_NATIVE_DMP;
    GLenum type           = GL_UNSIGNED_BYTE;

    u32 textureWidth      = bmpWidth;
    u32 textureHeight     = bmpHeight;

    u8* textureDataBuffer = NULL;
    textureDataBuffer = GetTextureDataFromBmpFileData(bmpWidth, bmpHeight, textureDataPtr,
                                                      textureWidth, textureHeight);
    s_RenderSystem.GenerateTexture(target, internalFormat, textureWidth,
                                   textureHeight, format, type, textureDataBuffer, textureId);
    s_AppHeap.Free(reinterpret_cast<void*>(textureDataBuffer));
    s_RenderSystem.SetColor(0.0f, 0.0f, 1.0f, 0.0f);

    f32 windowPositionX = 40.0f;
    f32 windowPositionY = 0.0f;

    f32 rectangleWidth = 320;
    f32 rectangleHeight = 240;

    s_RenderSystem.FillTexturedRectangle(textureId, windowPositionX, windowPositionY,
                                         rectangleWidth, rectangleHeight,
                                         static_cast<f32>(bmpWidth),
                                         static_cast<f32>(bmpHeight),
                                         static_cast<f32>(textureWidth),
                                         static_cast<f32>(textureHeight));
}
