﻿/*---------------------------------------------------------------------------*
  Project:  QRD
  File:     main.cpp

  Copyright (C) 2010-2014 Nintendo Co., Ltd.  All rights reserved.

  These coded instructions, statements, and computer programs contain
  proprietary information of Nintendo of America Inc. and/or Nintendo
  Company Ltd., and are protected by Federal copyright law.  They may
  not be disclosed to third parties or copied or duplicated in any form,
  in whole or in part, without the prior written consent of Nintendo.

  $Rev: $
 *---------------------------------------------------------------------------*/

#include <nn.h>
#include <nn/hid.h>
#include <nn/camera.h>
#include <nn/y2r.h>
#include <string.h>

#include <mw/qrd/qrd.h>
#include "GraphicsControl.h"

#define DISP_WIDTH          256    /**< Screen width */
#define DISP_HEIGHT         192    /**< Screen height */
#define CAMERA_WIDTH        640    /**< Width of a VGA-sized camera image */
#define CAMERA_HEIGHT       480    /**< Height of a VGA-sized camera image */
#define IMAGE_WIDTH         320    /**< Width of the trimmed image when using the VGA camera size */
#define IMAGE_HEIGHT        240    /**< Width of the trimmed image when using the VGA camera size */

#define Y2R_YUV_BYTES       2      /**< XUV byte count */
#define Y2R_RGB_BYTES       3      /**< RGB byte count */

/** @brief Application states. */
typedef enum
{
    QRDSTATUS_IDLE,         /**< Wait for decoding */
    QRDSTATUS_START,        /**< Start decoding */
    QRDSTATUS_DECODE,       /**< Decoding in progress */
    QRDSTATUS_PREVIEW       /**< Go into suspended status and display results of decoding */
} QRDStatus;

/** @brief Camera event states. */
enum
{
    EVENT_CAMERA_TRANSFER = 0,
    EVENT_CAMERA_ERROR,
    EVENT_CAMERA_NUM
};

GraphicsControl g_GraphicsControl;  /** Render operation class */

nn::fnd::ExpHeap s_AppHeap;         /** Heap used by the application */

using namespace mw::qrd;

// Buffer to recieve data converted to RGB by Y2R
u8 *g_RgbBuffer;
u8 *g_TmpRgbBuffer;
// Buffer to recieve camera capture data
u8 *g_YuvBuffer;
u8 *g_TmpYuvBuffer;

int g_top, g_left, g_right, g_bottom;

/// Size
size_t g_RgbBytes;
size_t g_CameraOutputSize;
size_t g_TransferUnit;
u32 g_Y2rOutputSize;

// Application status
static QRDStatus status_flag = QRDSTATUS_IDLE;

/**
 * @brief Displays guide rectangles on screen.
 *
 */
void StampGuide()
{
    int rate = 50;
    int cutWidth = CAMERA_WIDTH / 2;
    int cut_x = (IMAGE_WIDTH - cutWidth) / 2;
    int cutHeight = CAMERA_HEIGHT / 2;
    int cut_y = (IMAGE_HEIGHT - cutHeight) / 2;

    // Image starting point
    int start_x = (nn::gx::DISPLAY0_HEIGHT - (IMAGE_WIDTH * rate / 100)) / 2;
    // Guide starting point
    int rect_x = start_x + (cut_x * rate / 100);
    // Guide width
    int rectWidth = cutWidth * rate / 100;
    // Image starting point
    int start_y = (nn::gx::DISPLAY0_WIDTH - (IMAGE_HEIGHT * rate / 100)) / 2;
    // Guide starting point
    int rect_y = start_y + (cut_y * rate / 100);
    // Guide height
    int rectHeight = cutHeight * rate / 100;

    int top    = rect_y;
    int bottom = top + rectHeight;
    int left   = rect_x;
    int right  = left + rectWidth;

    int size = 20;
    f32 w = 2;
    f32 r = 1.0f;
    f32 g = 0.0f;
    f32 b = 0.0f;

    // upper left
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, left, top, left +size, top, w, r, g, b);
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, left, top, left, top +size, w, r, g, b);

    // upper right
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, right, top, right -10, top, w, r, g, b);
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, right, top, right, top +size, w, r, g, b);

    // lower left
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, left, bottom, left +size, bottom, w, r, g, b);
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, left, bottom, left, bottom -size, w, r, g, b);

    // upper right
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, right, bottom, right -size, bottom, w, r, g, b);
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, right, bottom, right, bottom -size, w, r, g, b);
}

/**
 * @brief Displays QR Code position information on screen.
 *
 * @param[in] top Position of the top of the QR Code
 * @param[in] left Position of the left of the QR Code
 * @param[in] right Position of the right of the QR Code
 * @param[in] bottom Position of the bottom of the QR Code
 */
void DrawLine(int top, int left, int right, int bottom)
{
    // Image starting point
    int rate = 50;
    int start_x = (nn::gx::DISPLAY0_HEIGHT - (CAMERA_WIDTH * rate / 100)) / 2;
    int startX = (left * rate / 100) + start_x;
    int startY = top  * rate / 100;
    int endX   = startX + (right - left) * rate / 100;
    int endY   = startY + (bottom - top) * rate / 100;
    f32 w = 2;
    f32 r = 0.0f;
    f32 g = 0.0f;
    f32 b = 1.0f;

    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, startX, startY, endX,   startY, w, r, g, b);
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, startX, startY, startX, endY  , w, r, g, b);
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, endX,   startY, endX,   endY  , w, r, g, b);
    g_GraphicsControl.DrawLine(NN_GX_DISPLAY0, startX, endY,   endX,   endY  , w, r, g, b);
}

/**
 * @brief Decode QR Code
 *
 * @param[in] decoder QR decoding library
 * @return Returns true if decoding is successful, returns false if decoding fails
 */
bool ExecuteDecode(QRDecoder *decoder)
{
    bool bRet = true;
    DecodeData res;
    ImageInfo info;
    int ret;

    memset(&res, 0, sizeof(res));
    memset(&info, 0, sizeof(info));

    info.width      = (unsigned short)CAMERA_WIDTH;
    info.height     = (unsigned short)CAMERA_HEIGHT;
    info.yuv422Data = (u32 *)g_YuvBuffer;

    // Set decode position
    info.qrPosition.top    = (CAMERA_HEIGHT - IMAGE_HEIGHT) / 2;
    info.qrPosition.left   = (CAMERA_WIDTH  - IMAGE_WIDTH)  / 2;
    info.qrPosition.right  = info.qrPosition.left + IMAGE_WIDTH;
    info.qrPosition.bottom = info.qrPosition.top  + IMAGE_HEIGHT;

    ret = decoder->ExecuteDecode(&res, &info);

    if (!ret)
    {
        // QR decoding succeeded
        // QR version
        NN_LOG("  QR Version: %d\n", res.version);
        // Size of content
        NN_LOG("  Contents size: %dbyte\n", res.size);
        // Content data
        NN_LOG("  Contents data: %s\n", res.data);
        // QR Code position
        NN_LOG("  qrPosition: top:%d, left:%d, right:%d, bottom:%d\n",
            res.qrPosition.top, res.qrPosition.left, res.qrPosition.right, res.qrPosition.bottom);
    	g_top = res.qrPosition.top;
    	g_left = res.qrPosition.left;
    	g_right = res.qrPosition.right;
    	g_bottom = res.qrPosition.bottom;
    }
    else
    {
        // QR decoding failed
        status_flag = QRDSTATUS_START;
        bRet = false;
    }

    return bRet;
}

/**
 * @brief Configures camera settings
 *
 * @param[in] decoder QR decoding library
 * @return Returns true if decoding is successful, returns false if decoding fails
 */
void CameraStart()
{
    memset(g_RgbBuffer, 0, CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_RGB_BYTES * sizeof(u8));
    memset(g_YuvBuffer, 0, CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_YUV_BYTES * sizeof(u8));

    // Specify the number of lines to be accumulated in the buffer
    s16 transferLine = nn::camera::GetMaxLines(CAMERA_WIDTH, CAMERA_HEIGHT);
    nn::camera::SetTransferLines(nn::camera::PORT_CAM1, transferLine, CAMERA_WIDTH, CAMERA_HEIGHT);

    // Change the frame rate
    nn::camera::SetFrameRate(nn::camera::SELECT_OUT1, nn::camera::FRAME_RATE_15); 

    // Change image size
    nn::camera::SetSize(nn::camera::SELECT_OUT1, nn::camera::SIZE_VGA, nn::camera::CONTEXT_A);

    // Put camera in startup state
    nn::camera::Activate(nn::camera::SELECT_OUT1);
    nn::camera::SetTrimming(nn::camera::PORT_CAM1, false);

    // Forcibly stop any conversion that is in progress
    nn::y2r::StopConversion();
    while(nn::y2r::IsBusyConversion()) // Wait because configuration cannot be done during Enable
    {
    }
    // Change input format
    nn::y2r::SetInputFormat(nn::y2r::INPUT_YUV422_BATCH);

    // Change output format
    nn::y2r::SetOutputFormat(nn::y2r::OUTPUT_RGB_24);

    g_RgbBytes = nn::y2r::GetOutputFormatBytes(nn::y2r::OUTPUT_RGB_24);

    // Change angle of rotation for output data
    nn::y2r::SetRotation(nn::y2r::ROTATION_NONE);

    // Change order of output data
    nn::y2r::SetBlockAlignment(nn::y2r::BLOCK_LINE);

    // Specify input data size
    nn::y2r::SetInputLineWidth(CAMERA_WIDTH);
    nn::y2r::SetInputLines(CAMERA_HEIGHT);

    // Set the conversion coefficients
    nn::y2r::SetStandardCoefficient(nn::y2r::COEFFICIENT_ITU_R_BT_601);

    // Calculate number of loops from the size of the image to transfer
    g_TransferUnit = nn::camera::GetTransferBytes(nn::camera::PORT_CAM1);

    // Number of bytes for 1 frame
    g_CameraOutputSize = nn::camera::GetFrameBytes(CAMERA_WIDTH, CAMERA_HEIGHT);
    g_Y2rOutputSize = nn::y2r::GetOutputImageSize(CAMERA_WIDTH, CAMERA_HEIGHT,
                                                  nn::y2r::OUTPUT_RGB_24);
}

/**
 * @brief A thread for camera error/data transfer processing
 *
 */
void CameraThreadFunc( uptr param )
{
    nn::os::Event  **events  = reinterpret_cast<nn::os::Event**>(param);

    // Configure the first transfer
    nn::camera::SetReceiving(events[EVENT_CAMERA_TRANSFER],
                             reinterpret_cast<void*>(g_TmpYuvBuffer),
                             nn::camera::PORT_CAM1, g_CameraOutputSize, g_TransferUnit);
    while(1)
    {
        s32 num = nn::os::WaitObject::WaitAny(reinterpret_cast<nn::os::WaitObject**>(events),
                                              EVENT_CAMERA_NUM);
        if (num == EVENT_CAMERA_TRANSFER)
        {
#if 0
            nn::camera::SetReceiving(events[EVENT_CAMERA_TRANSFER],
                                     reinterpret_cast<void*>(g_TmpYuvBuffer),
                                     nn::camera::PORT_CAM1, g_CameraOutputSize, g_TransferUnit);

            // Convert received YUV data with YUVtoRGB.
            nn::y2r::SetSendingYuv(reinterpret_cast<void*>(g_TmpYuvBuffer),
                                   g_CameraOutputSize, CAMERA_WIDTH * Y2R_YUV_BYTES);
            nn::y2r::SetReceiving(reinterpret_cast<void*>(g_RgbBuffer),
                                  g_Y2rOutputSize, CAMERA_WIDTH * 8* g_RgbBytes);
            nn::y2r::StopConversion();
            while(nn::y2r::IsBusyConversion()) // Wait because configuration cannot be done during Enable
            {
            }
			nn::Result res = nn::y2r::StartConversion();
            if (res.IsSuccess()) {
				memcpy(g_TmpRgbBuffer, g_RgbBuffer, CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_RGB_BYTES);
			}

            if (QRDSTATUS_START == status_flag)
            {
                status_flag = QRDSTATUS_DECODE;
                memcpy(g_YuvBuffer, g_TmpYuvBuffer, g_CameraOutputSize);
            }
#else
            if (QRDSTATUS_START == status_flag) {
                status_flag = QRDSTATUS_DECODE;
            }
			memcpy(g_YuvBuffer, g_TmpYuvBuffer, g_CameraOutputSize);
			nn::camera::SetReceiving(events[EVENT_CAMERA_TRANSFER],
                                     reinterpret_cast<void*>(g_TmpYuvBuffer),
                                     nn::camera::PORT_CAM1, g_CameraOutputSize, g_TransferUnit);

            nn::y2r::SetSendingYuv(reinterpret_cast<void*>(g_YuvBuffer),
                                   g_CameraOutputSize, CAMERA_WIDTH * Y2R_YUV_BYTES);
            nn::y2r::SetReceiving(reinterpret_cast<void*>(g_RgbBuffer),
                                  g_Y2rOutputSize, CAMERA_WIDTH * 8* g_RgbBytes);
            nn::y2r::StopConversion();
            while(nn::y2r::IsBusyConversion()) // Wait because configuration cannot be done during Enable
            {
				nn::os::Thread::Sleep(nn::fnd::TimeSpan::FromMicroSeconds(100));
            }
			nn::Result res = nn::y2r::StartConversion();
            if (res.IsSuccess()) {
				memcpy(g_TmpRgbBuffer, g_RgbBuffer, CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_RGB_BYTES);
			}
#endif
        }
        else if (num == EVENT_CAMERA_ERROR)
        {
            // Restart capture
            nn::camera::StopCapture(nn::camera::PORT_CAM1);
            nn::camera::ClearBuffer(nn::camera::PORT_CAM1);

            nn::camera::SetReceiving(events[EVENT_CAMERA_TRANSFER],
                                     reinterpret_cast<void*>(g_TmpYuvBuffer),
                                     nn::camera::PORT_CAM1, g_CameraOutputSize, g_TransferUnit);
            nn::camera::StartCapture(nn::camera::PORT_CAM1);
        }
    }
}

/**
 * @brief Create a thread for processing camera VSyncs
 *
 */
void CameraVsyncThreadFunc( uptr param )
{
    nn::os::Event  *vsyncEvent  = reinterpret_cast<nn::os::Event*>(param);

    while(1)
    {
        vsyncEvent->Wait();
    }
}

/**
 * @brief The Main function
 *
 */
void nnMain()
{
    bool bExit = false;     // Termination flag
    nn::Result result;      // Processing result
    u32 mem_size = 0;       // Buffer size for QR decoding working memory
    u32 *heap = NULL;       // Buffer for QR decoding working memory

    QRDecoder *decoder = NULL;

    // Initialize os
    // nn::os::Initialize();

    // Allocate heap
    s_AppHeap.Initialize(nn::os::GetDeviceMemoryAddress(),
						 nn::os::GetDeviceMemorySize(),
						 nn::os::ALLOCATE_OPTION_LINEAR);

	// Initialize the applet
	nn::applet::CTR::Enable();

    // Initialize graphics
    g_GraphicsControl.InitializeGraphics();

	// Allocate image work memory
	g_RgbBuffer = reinterpret_cast<u8 *>(s_AppHeap.Allocate(CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_RGB_BYTES));
    if (g_RgbBuffer == NULL)
    {
        NN_PANIC("ExpHeap.Allocate Error\n\n");
    }
	g_TmpRgbBuffer = reinterpret_cast<u8 *>(s_AppHeap.Allocate(CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_RGB_BYTES));
    if (g_TmpRgbBuffer == NULL)
    {
        NN_PANIC("ExpHeap.Allocate Error\n\n");
    }
	g_YuvBuffer = reinterpret_cast<u8 *>(s_AppHeap.Allocate(CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_YUV_BYTES));
    if (g_YuvBuffer == NULL)
    {
        NN_PANIC("ExpHeap.Allocate Error\n\n");
    }
	g_TmpYuvBuffer = reinterpret_cast<u8 *>(s_AppHeap.Allocate(CAMERA_WIDTH * CAMERA_HEIGHT * Y2R_YUV_BYTES));
    if (g_TmpYuvBuffer == NULL)
    {
        NN_PANIC("ExpHeap.Allocate Error\n\n");
    }

    // Initialize CAMERA
    nn::camera::Initialize();

    // Initialize Y2R
    nn::y2r::Initialize();

    nn::os::Event errorIntr, vsyncIntr;

    // Get interrupt handles
    nn::camera::GetBufferErrorInterruptEvent(&errorIntr, nn::camera::PORT_CAM1);
    nn::camera::GetVsyncInterruptEvent(&vsyncIntr, nn::camera::PORT_CAM1);

    // Camera-related settings
    CameraStart();

    nn::os::Event transEndIntr;
    // Get interrupt handles
    nn::y2r::GetTransferEndEvent(&transEndIntr);

    // Initialize hid
    result = nn::hid::Initialize();
    NN_UTIL_PANIC_IF_FAILED(result);

    // Prepare pad
    nn::hid::PadReader padReader;
    nn::hid::PadStatus padStatus;

    // Create thread for camera error/data transfer processing
    nn::os::Event dataTransferIntr;

    nn::os::Thread cameraThread;
    nn::os::Event *events[EVENT_CAMERA_NUM] = {&dataTransferIntr, &errorIntr};
    cameraThread.StartUsingAutoStack(
        CameraThreadFunc,
        reinterpret_cast<uptr>(events),
        4096,
        nn::os::Thread::GetCurrentPriority()-2 );

    // Create a thread for processing camera VSyncs
    nn::os::Thread cameraVsyncThread;
    cameraVsyncThread.StartUsingAutoStack(
        CameraVsyncThreadFunc,
        reinterpret_cast<uptr>(&vsyncIntr),
        4096,
        nn::os::Thread::GetCurrentPriority()-1 );

	// Get buffer size for QR decoding working memory
    mem_size = QRDecoder::GetDecodeBufferSize(CAMERA_WIDTH, CAMERA_HEIGHT);
    if (mem_size == 0)
    {
        NN_PANIC("failed QRDecoder::GetDecodeBufferSize().\n");
    }

    // Prepare the work buffer required for QR decoding
    heap = reinterpret_cast<u32*>(s_AppHeap.Allocate(mem_size));
    if (heap == NULL)
    {
        NN_PANIC("ExpHeap.Allocate Error\n\n");
    }

    // Initialize QR decoding
    decoder = new QRDecoder;
    if (!decoder)
    {
        NN_PANIC("Could not initialize QRD library...\n");
    }
    decoder->Initialize(heap, mem_size);

    // Start capture
    nn::camera::ClearBuffer(nn::camera::PORT_CAM1);
    nn::camera::StartCapture(nn::camera::PORT_CAM1);

    NN_LOG("\nBUTTON_A : Start QRDecode\n");
	
	bool res_draw = false;

    while(!bExit)
    {
        // Get input
        padReader.ReadLatest(&padStatus);

        transEndIntr.Wait();
        {

            if (QRDSTATUS_IDLE == status_flag || QRDSTATUS_START == status_flag)
            {
                // Initialize drawing.
                g_GraphicsControl.ClearDisplay(NN_GX_DISPLAY0);
                // Render camera
                g_GraphicsControl.DrawCamera(g_TmpRgbBuffer, CAMERA_WIDTH, CAMERA_HEIGHT);
                // Display decode position's rectangle specification in guide
                StampGuide();
            }
        	
        	if (QRDSTATUS_PREVIEW == status_flag && res_draw) {
        		res_draw = false;
                // Initialize drawing.
                g_GraphicsControl.ClearDisplay(NN_GX_DISPLAY0);
                // Render camera
                g_GraphicsControl.DrawCamera(g_TmpRgbBuffer, CAMERA_WIDTH, CAMERA_HEIGHT);
                // Display decode position's rectangle specification in guide
                StampGuide();

	        	// Render image position
     		    DrawLine(g_top, g_left, g_right, g_bottom);
        	}

            switch (status_flag)
            {
            case QRDSTATUS_IDLE:
                if (padStatus.trigger & nn::hid::BUTTON_A)
                {
                    status_flag = QRDSTATUS_START;
                }
                break;
            case QRDSTATUS_START:
                // Start QR decoding
                // Maintain the image data to decode from the camera buffer
                break;
            case QRDSTATUS_DECODE:
                // Decode QR Code
                if (ExecuteDecode(decoder))
                {
                    // Decoding succeeded
                    status_flag = QRDSTATUS_PREVIEW;
                	res_draw = true;
                }
                else
                {
                    // Failed to decode
                    status_flag = QRDSTATUS_START;
                }
                break;
            case QRDSTATUS_PREVIEW:
                if (padStatus.trigger & nn::hid::BUTTON_A)
                {
                    // Restart QR decoding
                    status_flag = QRDSTATUS_IDLE;
                    NN_LOG("\nBUTTON_A : Start QRDecode\n");
                }
                break;
            }
            // Display rendering results
            g_GraphicsControl.SwapBuffer(NN_GX_DISPLAY0, false);
            g_GraphicsControl.ClearDisplay(NN_GX_DISPLAY1);
            g_GraphicsControl.SwapBuffer();
        }
    }

    // Graphics termination processing
    g_GraphicsControl.FinalizeGraphics();

    // Finalize camera
    StopCapture(nn::camera::PORT_CAM1);
    ClearBuffer(nn::camera::PORT_CAM1);
    nn::y2r::StopConversion();
    nn::y2r::Finalize();
    nn::camera::Finalize();

    // End QR decoding
    decoder->Finalize();

    // Destroy heap
    if (NULL != heap)
    {
        s_AppHeap.Free(heap);
    }

    if (NULL != g_TmpYuvBuffer)
    {
        s_AppHeap.Free(g_TmpYuvBuffer);
    }
    if (NULL != g_YuvBuffer)
    {
        s_AppHeap.Free(g_YuvBuffer);
    }
    if (NULL != g_TmpRgbBuffer)
    {
        s_AppHeap.Free(g_TmpRgbBuffer);
    }
    if (NULL != g_RgbBuffer)
    {
        s_AppHeap.Free(g_RgbBuffer);
    }
    s_AppHeap.Finalize();
}
